"""
A short Python script
=====================

This demonstrates an example ``.py`` file that is not executed when gallery is
generated (see :ref:`build_pattern`) but nevertheless gets included as an
example. Note that no output is capture as this file is not executed.
"""

# Code source: Óscar Nájera
# License: BSD 3 clause
print([i for i in range(10)])
